//
// jja: swiss army knife for chess file formats
// src/pgn.rs: Portable Game Notation utilities
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::fmt::{self, Display, Formatter};
use std::fs::File;
use std::io::{stdout, Read, StdoutLock};

use indicatif::ProgressBar;

use pgcopy::Encoder;

use pgn_reader::{BufferedReader, RawHeader, SanPlus, Skip, Visitor};

use shakmaty::fen::{Epd, Fen};
use shakmaty::{CastlingMode, Chess, EnPassantMode, Position};

use crate::chess::serialize_chess;
use crate::hash::zobrist_hash;
use crate::system::get_progress_bar;
use crate::tr;

/// Output formats for `pgn_dump` function.
#[derive(Debug, Copy, Clone, Eq, PartialEq)]
pub enum OutputFormat {
    /// Represents the PosgreSQL binary output format.
    Binary,
    /// Represents the CSV output format.
    Csv,
    /// Represents the JSON output format.
    Json,
}

impl Display for OutputFormat {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        match self {
            OutputFormat::Binary => write!(f, "BIN"),
            OutputFormat::Csv => write!(f, "CSV"),
            OutputFormat::Json => write!(f, "JSON"),
        }
    }
}

/// `PositionTracker` is a struct that represents a chess position and tracks its validity.
///
/// # Fields
/// * `position`: A `Chess` struct representing the current chess position.
/// * `init`: A boolean that indicates whether the initial position of the game was processed.
/// * `valid`: A boolean that indicates whether the position is valid or not.
/// * `format`: A `Option<OutputFormat>` representing the optional output format.
struct PositionTracker<'a> {
    position: Chess,
    init: bool,
    valid: bool,
    format: Option<OutputFormat>,
    progress_bar: Option<&'a ProgressBar>,
    postgres_enc: Option<Encoder<StdoutLock<'a>>>,
}

impl<'a> PositionTracker<'a> {
    /// Constructs a new `PositionTracker` with the default chess position and sets the `valid`
    /// field to `true`.
    ///
    /// # Arguments
    ///
    /// - `format: Option<OutputFormat>`: Optional output format.
    /// - `progress_bar: Option<&ProgressBar>`: Optional progress bar.
    ///
    /// # Returns
    /// A new instance of `PositionTracker`.
    fn new(
        format: Option<OutputFormat>,
        encoder: Option<Encoder<StdoutLock<'a>>>,
        progress_bar: Option<&'a ProgressBar>,
    ) -> Self {
        PositionTracker {
            format,
            progress_bar,
            postgres_enc: encoder,
            position: Chess::default(),
            init: false,
            valid: true,
        }
    }

    fn write_header(&mut self) -> Result<(), Box<dyn std::error::Error>> {
        if let Some(encoder) = self.postgres_enc.as_mut() {
            encoder.write_header()?;
        }

        Ok(())
    }

    fn write_trailer(&mut self) -> Result<(), Box<dyn std::error::Error>> {
        if let Some(encoder) = self.postgres_enc.as_mut() {
            encoder.write_trailer()?;
        }

        Ok(())
    }

    fn dump_position(&mut self) {
        if let Some(format) = self.format {
            let key = zobrist_hash(&self.position);
            let vec = serialize_chess(&self.position);

            match format {
                OutputFormat::Binary => {
                    // SAFETY: output format guarantees postgres_enc is Some.
                    let enc = unsafe { self.postgres_enc.as_mut().unwrap_unchecked() };
                    enc.write_tuple(6).expect("pgcopy write tuple");
                    enc.write_bigint(key as i64).expect("pgcopy write key");
                    for item in &vec {
                        enc.write_bigint(*item as i64)
                            .expect("pgcopy write array item");
                    }
                }
                OutputFormat::Csv => {
                    println!(
                        "{},{},{},{},{},{}",
                        key as i64,
                        vec[0] as i64,
                        vec[1] as i64,
                        vec[2] as i64,
                        vec[3] as i64,
                        vec[4] as i64
                    );
                }
                OutputFormat::Json => {
                    println!(
                        "[{},{},{},{},{},{}]",
                        key, vec[0], vec[1], vec[2], vec[3], vec[4]
                    );
                }
            }

            if let Some(progress_bar) = &self.progress_bar {
                progress_bar.inc(1);
            }
        }
    }
}

/// Implementing the `Visitor` trait for the `PositionTracker` struct.
/// This allows the struct to handle chess games represented in SAN notation,
/// including playing moves, handling variations, and ending the game.
///
/// The `Result` type is set to `Chess`.
impl Visitor for PositionTracker<'_> {
    type Result = Chess;

    fn header(&mut self, key: &[u8], value: RawHeader<'_>) {
        if key != b"FEN" {
            return;
        }

        // Support games from a non-standard starting position.
        let fen = match Fen::from_ascii(value.as_bytes()) {
            Ok(fen) => fen,
            Err(err) => {
                eprintln!(
                    "{}",
                    tr!(
                        "Skipping invalid FEN header: {} ({}).",
                        err,
                        format!("{:?}", value)
                    )
                );
                self.valid = false;
                return;
            }
        };

        self.position = match fen.into_position(CastlingMode::Chess960) {
            Ok(pos) => pos,
            Err(err) => {
                eprintln!(
                    "{}",
                    tr!(
                        "Skipping illegal FEN header: {} ({}).",
                        err,
                        format!("{:?}", value)
                    )
                );
                self.valid = false;
                return;
            }
        };
        // Given multiple FEN headers (which is super-rare and formally
        // broken PGN), we process all of them for dump, and use the
        // last one for position tracker.
        self.dump_position();
        self.init = true;
    }

    fn end_headers(&mut self) -> Skip {
        if !self.init {
            self.dump_position();
            self.init = true;
        }
        Skip(!self.valid)
    }

    /// Processes a move in SAN notation and updates the position.
    ///
    /// If the position is invalid, this function does nothing.
    ///
    /// # Parameters
    /// * `san_plus`: A `SanPlus` containing the move in SAN notation.
    fn san(&mut self, san_plus: SanPlus) {
        if !self.valid {
            return;
        }

        let san = san_plus.san;
        let mov = match san.to_move(&self.position) {
            Ok(mov) => mov,
            Err(err) => {
                let epd = format!(
                    "{}",
                    Epd::from_position(self.position.clone(), EnPassantMode::PseudoLegal)
                );
                eprintln!(
                    "{}",
                    tr!("illegal SAN move `{}' in position `{}': {}", san, epd, err)
                );
                self.valid = false;
                return;
            }
        };

        self.position.play_unchecked(&mov);
        self.dump_position();
    }

    /// Handles the beginning of a variation.
    ///
    /// The implementation always returns `Skip(true)`, which causes the parser
    /// to stay in the mainline and ignore the variation.
    ///
    /// # Returns
    /// A `Skip` instance indicating whether to skip the variation or not.
    fn begin_variation(&mut self) -> Skip {
        Skip(true) // stay in the mainline
    }

    /// Handles the end of the game and resets the position.
    ///
    /// This function clones the current position and then resets the internal
    /// position to the default chess position.
    ///
    /// # Returns
    /// The `Chess` instance representing the final position of
    fn end_game(&mut self) -> Self::Result {
        self.valid = true;
        std::mem::take(&mut self.position)
    }
}

/// Converts a PGN string into an EPD string representing the final position.
///
/// This function takes a PGN string as input, processes the moves, and then
/// returns a string containing the final position in EPD format.
///
/// # Parameters
/// * `pgn`: A `&str` containing the PGN of the chess game.
///
/// # Returns
/// A `String` containing the final position of the game in EPD format.
pub fn pgn2epd(pgn: &str) -> String {
    let mut reader = BufferedReader::new_cursor(pgn);

    let mut tracker: PositionTracker<'_> = PositionTracker::new(None, None, None);
    let position = reader
        .read_game(&mut tracker)
        .expect("pgn")
        .expect("invalid pgn argument");

    format!(
        "{}",
        Epd::from_position(position, EnPassantMode::PseudoLegal)
    )
}

/// Converts a PGN file into a JSON stream of arrays which is a 6-sized array, the first element is
/// the Zobrist hash, the second to sixth elements are the serialized position as returned by
/// `jja::chess::serialize_chess` function or into a stream of CSV arrays which is again a 6-sized
/// array with the exact same structure as the JSON array except in CSV format the numbers are
/// cast into signed 64-bit numbers, whereas in JSON output they're unsigned 64-bit numbers.
/// Compressed PGN files are supported.
pub fn pgn_dump(file_name: &str, format: OutputFormat) -> Result<(), Box<dyn std::error::Error>> {
    eprintln!(
        "{}",
        tr!(
            "Dumping all positions in PGN file `{}' in format {} to standard output...",
            file_name,
            format
        )
    );

    let file = match File::open(file_name) {
        Ok(file) => file,
        Err(err) => {
            return Err(tr!("Failed to open PGN file `{}': {}", file_name, err).into());
        }
    };

    let file_ext = std::path::Path::new(file_name).extension();
    let uncompressed: Box<dyn Read + Send> =
        if file_ext.map_or(false, |ext| ext.eq_ignore_ascii_case("zst")) {
            Box::new(match zstd::Decoder::new(file) {
                Ok(decoder) => decoder,
                Err(err) => {
                    return Err(tr!("Failed to open PGN file `{}': {}", file_name, err).into());
                }
            })
        } else if file_ext.map_or(false, |ext| ext.eq_ignore_ascii_case("bz2")) {
            Box::new(bzip2::read::MultiBzDecoder::new(file))
        } else if file_ext.map_or(false, |ext| ext.eq_ignore_ascii_case("xz")) {
            Box::new(xz2::read::XzDecoder::new(file))
        } else if file_ext.map_or(false, |ext| ext.eq_ignore_ascii_case("gz")) {
            Box::new(flate2::read::GzDecoder::new(file))
        } else if file_ext.map_or(false, |ext| ext.eq_ignore_ascii_case("lz4")) {
            Box::new(match lz4::Decoder::new(file) {
                Ok(decoder) => decoder,
                Err(err) => {
                    return Err(tr!("Failed to open PGN file `{}': {}", file_name, err).into());
                }
            })
        } else {
            Box::new(file)
        };

    let progress_bar = get_progress_bar(0)?;
    progress_bar.set_message(tr!("Dumping:"));

    let encoder: Option<Encoder<StdoutLock>> = match format {
        OutputFormat::Binary => Some(Encoder::new(stdout().lock())),
        _ => None,
    };

    let mut tracker = PositionTracker::new(Some(format), encoder, Some(&progress_bar));
    tracker.write_header()?;
    BufferedReader::new(uncompressed).read_all(&mut tracker)?;
    tracker.write_trailer()?;

    let count = progress_bar.position();
    progress_bar.finish_and_clear();

    eprintln!(
        "{}",
        tr!(
            "Successfully dumped {} positions from PGN file `{} in format {} to standard output.",
            count,
            file_name,
            format
        )
    );
    Ok(())
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_pgn2epd() {
        let pgn = "1. e4 e5 2. Nf3 Nc6 3. Bb5";
        let epd = pgn2epd(pgn);
        assert_eq!(
            epd,
            "r1bqkbnr/pppp1ppp/2n5/1B2p3/4P3/5N2/PPPP1PPP/RNBQK2R b KQkq -"
        );
    }
}
